import pandas as pd
import sys
from datetime import datetime

class wsNGrams:
    words = []
    counts_df = pd.read_csv('matrix.csv')
    counts=counts_df.to_numpy()
    dates = []
    areColors = False
    colors = None
    nDates = len(counts_df.columns) - 1
    nWords = len(counts_df) - 1
    topN = sys.maxsize
    maxCount = 64

    def __init__(self, fName, startDateStr, endDateStr, topN):
        counts_df = pd.read_csv('matrix.csv')
        test = counts_df.reset_index(drop=True)
        test = test.iloc[0:, 0]
        test = test.dropna()
        words_list = list(test)
        for item in words_list:
            self.words.append(item)

        dates_list = list(counts_df.columns.values)[1:]
        for item in dates_list:
            item = datetime.strptime(item, '%Y-%m')
            item = item.date()
            self.dates.append(item)


import sys
import pygame                     # The PyGame 'sub-framework'
from .colorer import *            # Module for generating word colours
import sys
sys.path.append('./framework/')
from .framework import *
from pygame.locals import * # Some parts of the PyGame framework to import locally


try: # Try to import advanced freetype font module 
    import pygame.freetype as freetype
except ImportError:
    print ("No FreeType support compiled")
    sys.exit ()

class wsWordObj(freetype.Font):

	padding = 30; # The number of pixels around the text of the word YEARLY
			# to give a little better visual clash protection
	#@TODO Create an option for this that can be import from a
	#	settings file and to scale with screen size.

	

	def __init__(self, string, hues):
		super(wsWordObj, self).__init__(None)
	
		self.colorer = wsColorer(hues) # Initialize the colourer
	
		self.string = string # Assign the string
		
		# Determine aspect ratio using a big size 
		#  to avoid rounding errors
		self.size = 100 
		self.fgcolor = list(self.colorer.getColor())
		self.render = self.render(self.string)[0]		
		
		# Calculate aspect ratio of textbox with padding
		self.boxSize = self.get_rect(self.string)[2:4] # Box size (px)
		self.paddedAR = (self.boxSize[1] + self.padding) / (float(self.boxSize[0]  + self.padding))
		self.paddedWScale = (self.boxSize[0] - self.padding) / float(self.boxSize[0])
		self.paddedHScale = (self.boxSize[1] - self.padding) / float(self.boxSize[1])
		
	def Draw(self, screen, pos):
	
		# Draw the word in a box of width of newSize(x,y) (px)	
		newSize = list((0,0))
		newSize[0] = int(self.boxSize[0] * self.paddedWScale)
		newSize[1] = int(self.boxSize[1] * self.paddedHScale) 
		scaledRender = pygame.transform.smoothscale(self.render, newSize)
		pos = list(pos)
		pos[0] = pos[0] - (newSize[0] / 2)
		pos[1] = pos[1] - (newSize[1] / 2)			
		screen.blit(scaledRender, pos)




import random
import colorsys

class wsColorer():
	
	hues = None
	rgb = None
	
	def __init__(self, hues):
	
		# Initialize two hues that all words will be coloured with
		self.hues = [-1, -1]
		
		if len(hues) == 1:
			self.hues[0] = hues[0]
			self.hues[1] = hues[0]
			return
		elif len(hues) == 3:
			self.rgb = hues
			return
		
		if hues[0] == -1:
			self.hues[0] = random.uniform(0,1)
		else:
			self.hues[0] = hues[0]/255.0
			
		if hues[1] == -1:
			self.hues[1] = random.uniform(0,1)
		else:
			self.hues[1] = hues[1]/255.0
		
	def getColor(self):
	
		if self.rgb == None:
			# Choose one of the two different hues, then select a
			#  saturation and a value that works well with the background
			#  color. Then convert to 0-255 RGB values
			hue = self.hues[random.randint(0,1)]
			sat = random.uniform(0.1,0.9)
			val = random.uniform(0.5,0.9)
		
			rgb = list(colorsys.hsv_to_rgb(hue, sat, val))
			rgb[0] = rgb[0] * 255
			rgb[1] = rgb[1] * 255
			rgb[2] = rgb[2] * 255
			
		else:
			rgb = self.rgb
		
		return rgb


import os, getopt  # For system operations and argument parsing
import re
from .wsWordObj_monthly import *
from .NGrams_monthly import wsNGrams
import sys
sys.path.append('./framework/')
sys.path.append('./postprocessing/frames/')
from .framework.framework import Framework
from .framework.framework import *
from pygame.locals import * # Some parts of the PyGame framework to import locally



class WordSwarm(Framework):
	# Settings
	name = ""
	saveFrames = True
	csvName = 'matrix.csv'
	saveFolder = './postprocessing/frames/'

	frameN = 0;
	shapes = [];
	fixtures = [];
	bodies = [];
	joints = [];
	wordObjs = []
	sun = []
	wordHue = (-1, -1)
	topN = sys.maxsize

	# Size of words (m)
	maxSize = 6
	minSize = 0.1

	# Sun strength
	frequency = 0.1
	damping = 2

	def __init__(self, argv):

		# Initialize pygame/pybox2d framework/wcd ..orld
		super(WordSwarm, self).__init__()
		self.world.gravity = (0, 0);
		startDateStr = None
		endDateStr = None

		# Parse arguments
		try:
			opts, args = getopt.getopt(argv,
									   "hst:i:d:m:c:b:e:n:", ["ifile="])
		except getopt.GetoptError:
			print
			'Invalid argument'
			print(argv)
			print('try:')
			print
			'WordSwarm.py -n <topN] -s -t <title> -i <inputfile> -d <outputFolder> -m <maxWordHeight> -c <HexHue1_HexHue2> -b <startDate YYYYMMDD> -e <endDate YYYYMMDD>'
			sys.exit(2)

		for opt, arg in opts:
			if opt == '-h':
				print
				'WordSwarm.py -n <topN] -s -t <title> -i <inputfile> -d <outputFolder> -m <maxWordHeight> -c <HexHue1_HexHue2> -b <startDate YYYYMMDD> -e <endDate YYYYMMDD>'
				sys.exit()
			elif opt in ("-i", "--ifile"):

				print('Reading csv: %s' % arg)
				self.csvName = arg

			elif opt in ("-d"):
				self.saveFrames = True
				self.saveFolder = arg
				print('Saving frames to %s' % self.saveFolder)

			elif opt in ("-s"):
				self.saveFrames = True
				print('Saving frames to %s' % self.saveFolder)

			elif opt in ("-t"):

				print('WordSwarm title: %s' % arg)
				self.name = arg

			elif opt in ("-m"):
				self.maxSize = int(arg)
				print('Max word height set to: %d' % self.maxSize)

			elif opt in ("-c"):
				self.wordHue = (int(arg[0:2], 16), int(arg[3:5], 16))
				print('Words will have hues %d or %d' % (self.wordHue[0], self.wordHue[1]))

			elif opt in ("-b"):
				startDateStr = arg
				print('Starting animation at %s' % startDateStr)

			elif opt in ("-e"):
				endDateStr = arg
				print('Ending animation on %s' % endDateStr)

			elif opt in ("-n"):
				self.topN = int(arg)
				print('Displaying only the first %d results in csv' % self.topN)

		# Create output directory if required
		if self.saveFrames:
			if not os.path.exists(self.saveFolder):
				os.makedirs(self.saveFolder)
			else:
				self.purge(self.saveFolder, '.*')

		# Create ngrams database
		self.nGrams = wsNGrams(self.csvName, startDateStr, endDateStr, self.topN)

		box_half_size = (0.5, 0.5)

		# Place word objects
		self.shapes = [None] * self.nGrams.nWords
		self.fixtures = [None] * self.nGrams.nWords
		self.bodies = [None] * self.nGrams.nWords
		self.joints = [None] * self.nGrams.nWords
		self.wordObjs = [None] * self.nGrams.nWords;
		for word_k in range(0, self.nGrams.nWords):

			# The centre of the universe
			self.sun.append(self.world.CreateStaticBody(
				position=b2Vec2(random.uniform(
					-45, 65), 0)));

			# Create word object
			if self.nGrams.areColors == 'hue':
				self.wordObjs[word_k] = wsWordObj(
					self.nGrams.words[word_k], [self.nGrams.colors[word_k]])
			elif self.nGrams.areColors == 'rgb':
				self.wordObjs[word_k] = wsWordObj(
					self.nGrams.words[word_k], self.nGrams.colors[word_k])
			else:
				self.wordObjs[word_k] = wsWordObj(
					self.nGrams.words[word_k], self.wordHue)

			# Create body for word
			self.bodies[word_k] = self.world.CreateDynamicBody(
				position=(random.uniform(-60, 70), random.uniform(-40, 40)))

			# Add fixture to body
			self.fixtures[word_k] = self.bodies[word_k].CreateFixture(
				b2FixtureDef(shape=b2PolygonShape(box=(
					box_half_size[0] / self.wordObjs[word_k].paddedAR, box_half_size[0]))))

			# Link word object to sun
			self.joints[word_k] = self.world.CreateJoint(
				b2DistanceJointDef(
					frequencyHz=self.frequency,
					dampingRatio=self.damping,
					bodyA=self.sun[word_k],
					bodyB=self.bodies[word_k],
					localAnchorA=(0, 0),
					localAnchorB=(0, 0),
					length=0.5))

	# Removes files from a directory matching a pattern
	def purge(self, dir, pattern):
		for f in os.listdir(dir):
			if re.search(pattern, f):
				os.remove(os.path.join(dir, f))

	# Converts a size in (m) to a size in (px)
	def convertWorld2Screen(self, size_m):

		# Generate faux coordinates in (m)
		coord_m1 = size_m
		coord_m2 = list(size_m)
		coord_m2[0] = -coord_m2[0]
		coord_m2[1] = -coord_m2[1]

		# Convert faux coordinates to (px)
		coord_px1 = self.renderer.to_screen(coord_m1)
		coord_px2 = self.renderer.to_screen(coord_m2)

		# Calculate size from new coordinates
		size_p = [abs(coord_px1[0] - coord_px2[0]),
				  abs(coord_px1[1] - coord_px2[1])]

		return size_p

	# Draw the date progress bar
	# @TODO Scale text and line weight with screen size
	def Draw_Date(self, date_k):
		months = ['', 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
				  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

		color = (255, 255, 255)
		dateTxt = freetype.Font(None)
		dateTxt.size = 24
		dateTxt.fgcolor = color

		top = (int(self.screen.get_height() * 0.075), int(self.screen.get_height() * 0.125))
		#top = (250, 180)
		#bot = (250, 910)
		bot = (int(self.screen.get_height() * 0.075), int(self.screen.get_height() * (1 - 0.125)))

		self.screen.blit(dateTxt.render(self.name)[0],
						 (self.screen.get_height() * 0.03, self.screen.get_height() * 0.03))

		pygame.draw.line(self.screen, color, top, bot, 4)
		pygame.draw.line(self.screen, color, (top[0] - 4, top[1]), (top[0] + 4, top[1]), 4)
		pygame.draw.line(self.screen, color, (bot[0] - 4, bot[1]), (bot[0] + 4, bot[1]), 4)

		self.screen.blit(dateTxt.render('%s %04d' % (
			months[self.nGrams.dates[0].month],
			self.nGrams.dates[0].year))[0],
						 (self.screen.get_height() * 0.03, top[1] - 40))

		self.screen.blit(dateTxt.render('%s %04d' % (
			months[self.nGrams.dates[-1].month],
			self.nGrams.dates[-1].year))[0],
						 (self.screen.get_height() * 0.03, bot[1] + 20))

		progress = ((self.nGrams.dates[date_k] - self.nGrams.dates[0]).total_seconds() /
					(self.nGrams.dates[-1] - self.nGrams.dates[0]).total_seconds())

		pos = (top[0] + 1, int((bot[1] - top[1]) * progress + top[1]))
		pygame.draw.circle(self.screen, color, pos, 8)

		self.screen.blit(dateTxt.render('%s %04d' % (
			months[self.nGrams.dates[date_k].month],
			self.nGrams.dates[date_k].year))[0],
						 (pos[0] + 20, pos[1] - 10))


	def Step(self, settings):
		self.screen.fill((0, 0, 0))
		self.frameN = self.frameN + 1;

		# Update ngram date every n-frames
		nFrames = 90
		date_k = int(self.frameN / nFrames) + 1  # Date moving towards
		phase = (self.frameN % nFrames) / float(nFrames)  # (0-1) way to new date

		# Stop if at the end of the dataset
		if date_k == self.nGrams.nDates:
			print('WordSwarm animation complete')
			print('Hope you enjoyed the show!')
			exit()

		dateTxt = freetype.Font(None)
		dateTxt.size = 10
		dateTxt.fgcolor = (255, 255, 255)

		self.Draw_Date(date_k)  # Draw the date

		# Update size of each word
		# @TODO There is a memory leak in creating the new bodies (it doesn't delete the old ones)
		for word_k in range(0, self.nGrams.nWords):

			# Calculate word sizes
			if phase == 1:
				newSize = (0, 0)

				newSize[1] = (self.maxSize - self.minSize) * (
							self.nGrams.counts[word_k][date_k - 1] / float(self.nGrams.maxCount)) + self.minSize

				newSize[0] = newSize[1] / self.wordObjs[word_k].paddedAR

				self.wordObjs[word_k].boxSize = self.convertWorld2Screen(newSize)
			else:
				newSize = list((0, 0))
				newSize[1] = (self.maxSize - self.minSize) * (
							self.nGrams.counts[word_k][date_k] / float(self.nGrams.maxCount)) + self.minSize
				newSize[0] = newSize[1] / self.wordObjs[word_k].paddedAR

				oldSize = list((0, 0))
				oldSize[1] = (self.maxSize - self.minSize) * (
							self.nGrams.counts[word_k][date_k - wsNGrams.nDates] / float(self.nGrams.maxCount)) + self.minSize
				oldSize[0] = oldSize[1] / self.wordObjs[word_k].paddedAR

				size = list((0, 0))
				size[0] = (newSize[0] - oldSize[0]) * phase + oldSize[0]
				size[1] = (newSize[1] - oldSize[1]) * phase + oldSize[1]

				self.wordObjs[word_k].boxSize = self.convertWorld2Screen(size)

			# Re-create fixture
			self.bodies[word_k].DestroyFixture(self.fixtures[word_k])
			self.fixtures[word_k] = None
			self.fixtures[word_k] = self.bodies[word_k].CreateFixture(
				b2FixtureDef(shape=b2PolygonShape(box=newSize)))

			# Redraw word in new shape
			pos = self.renderer.to_screen(self.bodies[word_k].position);
			self.wordObjs[word_k].Draw(self.screen, pos)

		# Save frames to create film
		if self.saveFrames:
			pygame.image.save(self.screen,
							  self.saveFolder + int(self.frameN).__format__('')
							  + '.png')

		Framework.Step(self, settings);


from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import QObject, SIGNAL, pyqtSlot
from PyQt4.QtGui import QTableWidget, QTableWidgetItem, QColor, QPixmap
from PyQt4.QtCore import Qt, QSettings
from .pyqt4_gui import Ui_MainWindow
from .framework import *
from time import time
from .settings import *
import string
import sys
import re

class Pyqt4Draw(object):

    MAX_TIMES=20
    axisScale = 0.4
    def __init__(self, test): 
        self.test=test
        self.window=self.test.window
        self.scene=self.window.scene
        self.view=self.window.graphicsView
        self.item_cache={}
        self.temp_items=[]
        self.status_font=QtGui.QFont("Times", 10, QtGui.QFont.Bold)
        self.font_spacing=QtGui.QFontMetrics(self.status_font).lineSpacing()
        self.draw_idx=0

    def StartDraw(self):
        for item in self.temp_items:
            self.scene.removeItem(item)
        self.temp_items=[]

    def EndDraw(self): pass

    def SetFlags(self, **kwargs): 
        """
        For compatibility with other debug drawing classes.
        """
        pass

    def DrawStringAt(self, x, y, str, color):
        item=QtGui.QGraphicsSimpleTextItem(str)
        #brush=QtGui.QBrush(QColor(*color))
        brush=QtGui.QBrush(QColor(255,255,255,255))
        item.setFont(self.status_font)
        item.setBrush(brush)
        item.setPos(self.view.mapToScene(x,y))
        item.scale(1./self.test._viewZoom, -1./self.test._viewZoom)
        self.temp_items.append(item)

        self.scene.addItem(item)

    def DrawPoint(self, p, size, color):
        """
        Draw a single point at point p given a pixel size and color.
        """
        self.DrawCircle(p, size/self.test.viewZoom, color, drawwidth=0)
        
    def DrawAABB(self, aabb, color):
        """
        Draw a wireframe around the AABB with the given color.
        """
        line1=self.scene.addLine(aabb.lowerBound.x, aabb.lowerBound.y, aabb.upperBound.x, aabb.lowerBound.y, 
                         pen=QtGui.QPen(QColor(*color.bytes)))
        line2=self.scene.addLine(aabb.upperBound.x, aabb.upperBound.y, aabb.lowerBound.x, aabb.upperBound.y,
                         pen=QtGui.QPen(QColor(*color.bytes)))
        self.temp_items.append(line1)
        self.temp_items.append(line2)

    def DrawSegment(self, p1, p2, color):
        """
        Draw the line segment from p1-p2 with the specified color.
        """
        line=self.scene.addLine(p1[0], p1[1], p2[0], p2[1], pen=QtGui.QPen(QColor(*color.bytes)))
        self.temp_items.append(line)

    def DrawTransform(self, xf):
        """
        Draw the transform xf on the screen
        """
        p1 = xf.position
        p2 = p1 + self.axisScale * xf.R.col1
        p3 = p1 + self.axisScale * xf.R.col2

        line1=self.scene.addLine(p1[0], p1[1], p2[0], p2[1], pen=QtGui.QPen(QColor(255,0,0)))
        line2=self.scene.addLine(p1[0], p1[1], p3[0], p3[1], pen=QtGui.QPen(QColor(0,255,0)))
        self.temp_items.append(line1)
        self.temp_items.append(line2)

    def DrawCircle(self, center, radius, color, drawwidth=1, shape=None):
        """
        Draw a wireframe circle given the center, radius, axis of orientation and color.
        Note that these functions 
        """
        border_color=[c*255 for c in color] + [255]
        pen  =QtGui.QPen(QtGui.QColor(*border_color))
        ellipse=self.scene.addEllipse(center[0]-radius, center[1]-radius, radius*2, radius*2, pen=pen)
        self.temp_items.append(ellipse)

    def DrawSolidCircle(self, center, radius, axis, color, shape=None):
        """
        Draw a solid circle given the center, radius, axis of orientation and color.
        """
        border_color=color.bytes + [255]
        inside_color=(color / 2).bytes + [127]
        brush=QtGui.QBrush(QtGui.QColor(*inside_color))
        pen  =QtGui.QPen(QtGui.QColor(*border_color))
        ellipse=self.scene.addEllipse(center[0]-radius, center[1]-radius, radius*2, radius*2, brush=brush, pen=pen)
        line=self.scene.addLine(center[0], center[1], (center[0]-radius*axis[0]), (center[1]-radius*axis[1]), pen=QtGui.QPen(QColor(255,0,0)))

        self.temp_items.append(ellipse)
        self.temp_items.append(line)

    def DrawPolygon(self, vertices, color, shape=None):
        """
        Draw a wireframe polygon given the world vertices vertices (tuples) with the specified color.
        """
        poly=QtGui.QPolygonF()
        pen=QtGui.QPen(QtGui.QColor(*color.bytes))

        for v in vertices:
            poly+=QtCore.QPointF(*v)

        item=self.scene.addPolygon(poly, pen=pen)
        self.temp_items.append(item)

    def DrawSolidPolygon(self, vertices, color, shape=None):
        """
        Draw a filled polygon given the world vertices vertices (tuples) with the specified color.
        """
        poly=QtGui.QPolygonF()
        border_color=color.bytes + [255]
        inside_color=(color / 2).bytes + [127]
        brush=QtGui.QBrush(QtGui.QColor(*inside_color))
        pen  =QtGui.QPen(QtGui.QColor(*border_color))

        for v in vertices:
            poly+=QtCore.QPointF(*v)

        item=self.scene.addPolygon(poly, brush=brush, pen=pen)
        self.temp_items.append(item)

    def DrawCircleShape(self, shape, transform, color, temporary=False):
        center=b2Mul(transform, shape.pos)
        radius=shape.radius
        axis=transform.R.col1

        border_color=color.bytes + [255]
        inside_color=(color / 2).bytes + [127]
        brush=QtGui.QBrush(QtGui.QColor(*inside_color))
        pen  =QtGui.QPen(QtGui.QColor(*border_color))
        ellipse=self.scene.addEllipse(-radius, -radius, radius*2, radius*2, brush=brush, pen=pen)
        line=self.scene.addLine(center[0], center[1], (center[0]-radius*axis[0]), (center[1]-radius*axis[1]), pen=QtGui.QPen(QColor(255,0,0)))
        ellipse.setPos(*center)
        ellipse.radius=radius

        if temporary:
            self.temp_items.append(ellipse)
            self.temp_items.append(line)
        else:
            self.item_cache[hash(shape)]=[ellipse, line]
        
    def DrawPolygonShape(self, shape, transform, color, temporary=False):
        poly=QtGui.QPolygonF()
        border_color=color.bytes + [255]
        inside_color=(color / 2).bytes + [127]
        brush=QtGui.QBrush(QtGui.QColor(*inside_color))
        pen  =QtGui.QPen(QtGui.QColor(*border_color))

        for v in shape.vertices:
            poly+=QtCore.QPointF(*v)

        item=self.scene.addPolygon(poly, brush=brush, pen=pen)
        item.setRotation(transform.angle * 180.0 / b2_pi)
        item.setPos(*transform.position)

        if temporary:
            self.temp_items.append(item)
        else:
            self.item_cache[hash(shape)]=[item]

    def _remove_from_cache(self, shape):
        items=self.item_cache[hash(shape)]

        del self.item_cache[hash(shape)]
        for item in items:
            self.scene.removeItem(item)
    
    def DrawShape(self, shape, transform, color, selected=False):
        """
        Draw any type of shape
        """
        cache_hit=False
        if hash(shape) in self.item_cache:
            cache_hit=True
            items=self.item_cache[hash(shape)]
            items[0].setRotation(transform.angle * 180.0 / b2_pi)
            if isinstance(shape, b2CircleShape):
                radius=shape.radius
                if items[0].radius == radius:
                    center=b2Mul(transform, shape.pos)
                    items[0].setPos(*center)
                    line=items[1]
                    axis=transform.R.col1
                    line.setLine(center[0], center[1], (center[0]-radius*axis[0]), (center[1]-radius*axis[1]))
                else:
                    self._remove_from_cache(shape)
                    cache_hit=False
            else:
                items[0].setPos(*transform.position)

            if not selected or cache_hit:
                return

        if selected:
            color=b2Color(1,1,1)
            temporary=True
        else:
            temporary=False

        if isinstance(shape, b2PolygonShape):
            self.DrawPolygonShape(shape, transform, color, temporary)
        elif isinstance(shape, b2EdgeShape):
            v1=b2Mul(transform, shape.vertex1)
            v2=b2Mul(transform, shape.vertex2)
            self.DrawSegment(v1, v2, color)
        elif isinstance(shape, b2CircleShape):
            self.DrawCircleShape(shape, transform, color, temporary)
        elif isinstance(shape, b2LoopShape):
            vertices=shape.vertices
            v1=b2Mul(transform, vertices[-1])
            for v2 in vertices:
                v2=b2Mul(transform, v2)
                self.DrawSegment(v1, v2, color)
                v1=v2

    def DrawJoint(self, joint):
        """
        Draw any type of joint
        """
        bodyA, bodyB=joint.bodyA, joint.bodyB
        xf1, xf2=bodyA.transform, bodyB.transform
        x1, x2=xf1.position, xf2.position
        p1, p2=joint.anchorA, joint.anchorB
        color=b2Color(0.5, 0.8, 0.8)
            
        if isinstance(joint, b2DistanceJoint):
            self.DrawSegment(p1, p2, color)
        elif isinstance(joint, b2PulleyJoint):
            s1, s2=joint.groundAnchorA, joint.groundAnchorB
            self.DrawSegment(s1, p1, color)
            self.DrawSegment(s2, p2, color)
            self.DrawSegment(s1, s2, color)

        elif isinstance(joint, b2MouseJoint):
            pass # don't draw it here
        else:
            self.DrawSegment(x1, p1, color)
            self.DrawSegment(p1, p2, color)
            self.DrawSegment(x2, p2, color)

    def ManualDraw(self):
        """
        This implements code normally present in the C++ version,
        which calls the callbacks that you see in this class (DrawSegment,
        DrawSolidCircle, etc.).
        
        This is implemented in Python as an example of how to do it, and also
        a test.
        """
        colors = {
            'active'    : b2Color(0.5, 0.5, 0.3),
            'static'    : b2Color(0.5, 0.9, 0.5), 
            'kinematic' : b2Color(0.5, 0.5, 0.9), 
            'asleep'    : b2Color(0.6, 0.6, 0.6), 
            'default'   : b2Color(0.9, 0.7, 0.7), 
        }

        settings=self.test.settings
        world=self.test.world
        if self.test.selected_shapebody:
            sel_shape, sel_body=self.test.selected_shapebody
        else:
            sel_shape=None

        if settings.drawShapes:
            for body in world.bodies:
                transform=body.transform
                for fixture in body.fixtures:
                    shape=fixture.shape

                    if not body.active: color=colors['active']
                    elif body.type==b2_staticBody: color=colors['static']
                    elif body.type==b2_kinematicBody: color=colors['kinematic']
                    elif not body.awake: color=colors['asleep']
                    else: color=colors['default']
                    
                    self.DrawShape(fixture.shape, transform, color, (sel_shape==shape))


        if settings.drawJoints:
            for joint in world.joints:
                self.DrawJoint(joint)

        # if settings.drawPairs
        #   pass

        if settings.drawAABBs:
            color=b2Color(0.9, 0.3, 0.9)
            cm=world.contactManager
            for body in world.bodies:
                if not body.active:
                    continue
                transform=body.transform
                for fixture in body.fixtures:
                    shape=fixture.shape
                    for childIndex in range(shape.childCount):
                        self.DrawAABB(shape.getAABB(transform, childIndex), color)

    def to_screen(self, point):
        """
        In here for compatibility with other frameworks.
        """
        return tuple(point)

class GraphicsScene (QtGui.QGraphicsScene):
    def __init__(self, test, parent=None):
        super(GraphicsScene, self).__init__(parent)
        self.test=test

    def keyPressEvent(self, event):
        self.test._Keyboard_Event(event.key(), down=True)

    def keyReleaseEvent(self, event):
        self.test._Keyboard_Event(event.key(), down=False)

    def mousePressEvent(self, event):
        pos=self.test.ConvertScreenToWorld(event.scenePos().x(), event.scenePos().y())
        if event.button()==Qt.RightButton:
            self.test.ShowProperties(pos)
        elif event.button()==Qt.LeftButton:
            if event.modifiers() == Qt.ShiftModifier:
                self.test.ShiftMouseDown(pos)
            else:
                self.test.MouseDown(pos)

    def mouseReleaseEvent(self, event):
        pos=event.scenePos().x(), event.scenePos().y()
        if event.button()==Qt.RightButton:
            self.test.MouseUp(pos)
        elif event.button()==Qt.LeftButton:
            self.test.MouseUp(pos)

    def mouseMoveEvent(self, event):
        pos=event.scenePos().x(), event.scenePos().y()
        self.test.MouseMove(self.test.ConvertScreenToWorld(*pos))
        QtGui.QGraphicsScene.mouseMoveEvent(self, event)
        
class MainWindow(QtGui.QMainWindow,Ui_MainWindow):
    def __init__(self, test, parent=None):
        QtGui.QMainWindow.__init__(self)
        self.setupUi(self)
        self.scene=GraphicsScene(test)
        self.test=test
        self.scene.setBackgroundBrush( QtGui.QBrush(QtGui.QColor(0,0,0)) )
        self.graphicsView.setScene(self.scene)
        self.graphicsView.scale(self.test.viewZoom, -self.test.viewZoom)
        self.reset_properties_list()
        self.restoreLayout()

        QObject.connect(self.mnuExit, SIGNAL("triggered()"), self.close)
        QObject.connect(self.mnuIncreaseFontSize, SIGNAL("triggered()"), lambda amount= 1.0 : self.setFontSize(app.font().pointSize()+amount))
        QObject.connect(self.mnuDecreaseFontSize, SIGNAL("triggered()"), lambda amount=-1.0 : self.setFontSize(app.font().pointSize()+amount))

        self.add_settings_widgets()

    def add_settings_widgets(self):
        self.settings_widgets={}

        gb=self.gbOptions # the options groupbox
        layout=QtGui.QVBoxLayout()
        gb.setLayout(layout)

        for text, variable in settings.checkboxes:
            if variable:
                widget=QtGui.QCheckBox('&' + text)
                QObject.connect(widget, SIGNAL("stateChanged(int)"),
                        lambda _, var=variable, widget=widget : setattr(self.test.settings, var, widget.isChecked()) )
                widget.setChecked(getattr(self.test.settings, variable))
                self.settings_widgets[variable]=widget
            else:
                widget=QtGui.QLabel(text)
                widget.setAlignment(Qt.AlignHCenter)

            layout.addWidget(widget)

        def update_slider(var, value, text, label):
            setattr(self.test.settings, var, value)
            label.setText('%s (%d)' % (text, value))

        for slider in settings.sliders:
            label=QtGui.QLabel(slider['text'])
            label.setAlignment(Qt.AlignHCenter)
            layout.addWidget(label)

            widget=QtGui.QScrollBar(Qt.Horizontal)
            widget.setRange(slider['min'], slider['max'])
            var=slider['name']
            QObject.connect(widget, SIGNAL("valueChanged(int)"), 
                lambda value, var=var, text=slider['text'], label=label: update_slider(var, value, text, label) )
            self.settings_widgets[var]=widget
            layout.addWidget(widget)

        self.update_widgets_from_settings()

    def update_widgets_from_settings(self, step_settings=None):
        if step_settings is None:
            step_settings=self.test.settings
        
        for var, widget in list(self.settings_widgets.items()):
            if isinstance(widget, QtGui.QCheckBox):
                widget.setChecked(getattr(step_settings, var))
            else:
                widget.setValue(getattr(step_settings, var))

        for slider in settings.sliders:
            var=slider['name']
            self.settings_widgets[var].setValue(getattr(step_settings, var))

    def reset_properties_list(self):
        self.twProperties.clear()
        self.twProperties.setRowCount(0)
        self.twProperties.setColumnCount(3)
        self.twProperties.verticalHeader().hide() # don't show numbers on left
        self.twProperties.setHorizontalHeaderLabels(['class', 'name', 'value'])
    
    def keyPressEvent(self, event):
        self.test._Keyboard_Event(event.key(), down=True)

    def keyReleaseEvent(self, event):
        self.test._Keyboard_Event(event.key(), down=False)

    @property
    def settings(self):
        return QtCore.QSettings("pybox2d", "Framework") 

    def setFontSize(self, size):
        """
        Update the global font size
        """
        if size <= 0.0:
            return

        global app
        font=app.font()
        font.setPointSize(size)
        app.setFont(font)

    def restoreLayout(self):
        """
        Restore the layout of each widget
        """
        settings=self.settings
        try:
            self.restoreGeometry(settings.value("geometry").toByteArray())
            self.restoreState(settings.value("windowState").toByteArray())
            size=settings.value('fontSize').toFloat()[0]
            self.setFontSize(size)
        except:
            pass

    def saveLayout(self):
        """
        Save the layout of each widget
        """
        settings=self.settings
        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("windowState", self.saveState())
        settings.setValue("fontSize", app.font().pointSize())

    def closeEvent(self, event):
        QtGui.QMainWindow.closeEvent(self, event)
        self.saveLayout()

app=None
class Pyqt4Framework(FrameworkBase):
    TEXTLINE_START=0
    def setup_keys(self):
        # Only basic keys are mapped for now: K_[a-z0-9], K_F[1-12] and K_COMMA.

        for letter in string.uppercase:
            setattr(Keys, 'K_'+letter.lower(), getattr(Qt, 'Key_%s' % letter))
        for i in range(0,10):
            setattr(Keys, 'K_%d'%i, getattr(Qt, 'Key_%d' % i))
        for i in range(1,13):
            setattr(Keys, 'K_F%d'%i, getattr(Qt, 'Key_F%d' % i))
        Keys.K_LEFT=Qt.Key_Left
        Keys.K_RIGHT=Qt.Key_Right
        Keys.K_UP=Qt.Key_Up
        Keys.K_DOWN=Qt.Key_Down
        Keys.K_HOME=Qt.Key_Home
        Keys.K_PAGEUP=Qt.Key_PageUp
        Keys.K_PAGEDOWN=Qt.Key_PageDown
        Keys.K_COMMA=Qt.Key_Comma
        Keys.K_SPACE=Qt.Key_Space

    def __reset(self):
        # Screen/rendering-related
        self._viewZoom         =10.0
        self._viewCenter       =None
        self._viewOffset       =None
        self.screenSize        =None
        self.textLine          =0
        self.font              =None
        self.fps               =0
        self.selected_shapebody=None, None

        # GUI-related
        self.window=None
        self.setup_keys()
        
    def __init__(self):
        super(Pyqt4Framework, self).__init__()
        
        if fwSettings.onlyInit: # testing mode doesn't initialize Pyqt4
            return
            
        global app
        app = QtGui.QApplication(sys.argv)

        self.__reset()
        print('Initializing Pyqt4 framework...')

        # Pyqt4 Initialization
        self.window=MainWindow(self)
        self.window.show()

        self.window.setWindowTitle( "Python Box2D Testbed - " + self.name)
        self.renderer = Pyqt4Draw(self)

        # Note that in this framework, we override the draw debug data routine
        # that occurs in Step(), and we implement the normal C++ code in Python.
        self.world.DrawDebugData = lambda: self.renderer.ManualDraw()
        self.screenSize = b2Vec2(0,0)
        self.viewCenter = (0,10.0*20.0)
        self.groundbody = self.world.CreateBody()

    def setCenter(self, value):
        """
        Updates the view offset based on the center of the screen.
        
        Tells the debug draw to update its values also.
        """
        self._viewCenter = b2Vec2( *value )
        self._viewOffset = self._viewCenter - self.screenSize/2
        self.window.graphicsView.centerOn(*self._viewCenter)
    
    def setZoom(self, zoom):
        self._viewZoom = zoom
        self.window.graphicsView.resetTransform()
        self.window.graphicsView.scale(self._viewZoom, -self._viewZoom)
        self.window.graphicsView.centerOn(*self._viewCenter)

    viewZoom   = property(lambda self: self._viewZoom, setZoom,
                           doc='Zoom factor for the display')
    viewCenter = property(lambda self: self._viewCenter, setCenter, 
                           doc='Screen center in camera coordinates')
    viewOffset = property(lambda self: self._viewOffset,
                           doc='The offset of the top-left corner of the screen')

    def run(self):
        """
        What would be the main loop is instead a call to 
        app.exec_() for the event-driven pyqt4.
        """
        global app
        self.step_timer = QtCore.QTimer()
        QObject.connect(self.step_timer, SIGNAL("timeout()"), lambda : self.SimulationLoop())
        QObject.connect(self.window.twProperties, SIGNAL('itemChanged(QTableWidgetItem*)'), self.prop_cell_changed)
        self.step_timer.start(int((1000.0/self.settings.hz)))

        app.exec_()
        
        self.step_timer.stop()
        print('Cleaning up...')
        self.world.contactListener=None
        self.world.destructionListener=None
        self.world.renderer=None
        self.world=None

    def _Keyboard_Event(self, key, down=True):
        """
        Internal keyboard event, don't override this.

        Checks for the initial keydown of the basic testbed keys. Passes the unused
        ones onto the test via the Keyboard() function.
        """
        if down:
            if key==Keys.K_z:       # Zoom in
                self.viewZoom = min(1.10 * self.viewZoom, 50.0)
            elif key==Keys.K_x:     # Zoom out
                self.viewZoom = max(0.9 * self.viewZoom, 0.02)
            elif key==Keys.K_SPACE: # Launch a bomb
                self.LaunchRandomBomb()
            else:              # Inform the test of the key press
                self.Keyboard(key)
        else:
            self.KeyboardUp(key)

    def CheckKeys(self):
        pass

    def _ShowProperties(self, obj):
        self.selected_shapebody=None, None

        class_=obj.__class__
        ignore_list=('thisown',)

        i=0
        twProperties=self.window.twProperties
        # Get all of the members of the class
        for prop in dir(class_):
            # If they're properties and not to be ignored, add them to the table widget
            if isinstance(getattr(class_, prop), property) and prop not in ignore_list:
                try:
                    value=getattr(obj, prop)
                except:
                    # Write-only?
                    continue

                widget=None

                # Attempt to determine whether it's read-only or not
                try:
                    setattr(obj, prop, value)
                except:
                    editable=False
                else:
                    editable=True

                # Increase the row count and insert the new item                
                twProperties.setRowCount(twProperties.rowCount()+1)
                i=twProperties.rowCount()-1
                self.item=QTableWidgetItem(class_.__name__)
                twProperties.setItem(i, 0, QTableWidgetItem(class_.__name__)) # class name
                twProperties.item(i, 0).setFlags(Qt.ItemIsEnabled)

                twProperties.setItem(i, 1, QtGui.QTableWidgetItem(prop))      # prop name
                twProperties.item(i, 1).setFlags(Qt.ItemIsEnabled)

                # and finally, the property values
                # booleans are checkboxes
                if isinstance(value, bool):
                    widget=QtGui.QCheckBox('')
                    QObject.connect(widget, SIGNAL('stateChanged(int)'), 
                                    lambda value, prop=prop: self.property_changed(prop, value==Qt.Checked))
                    if value:
                        widget.setCheckState(Qt.Checked)
                # ints, floats are spinboxes
                elif isinstance(value, (int, float)):
                    widget=QtGui.QDoubleSpinBox()
                    QObject.connect(widget, SIGNAL('valueChanged(double)'), 
                                    lambda value, prop=prop: self.property_changed(prop, value))
                    widget.setValue(value)
                # lists turn into -- lists
                elif isinstance(value, list):
                    widget=QtGui.QListWidget()
                    for entry in value:
                        widget.addItem(str(entry))
                    if value:
                        #sz=widget.item(0).sizeHint()
                        #print(sz, sz.width(), sz.height())
                        #sz.setHeight(sz.height()*2)
                        #widget.setMinimumSize(sz)
                        #widget.setMinimumSize(QtCore.QSize(1,60))
                        pass # TODO
                # vec2s will be shown as a textbox
                elif isinstance(value, b2Vec2):
                    value='(%.2f, %.2f)' % (tuple(value))
                else:
                    pass

                if widget:
                    twProperties.setCellWidget(i, 2, widget)
                    if hasattr(widget, 'setReadOnly'):
                        widget.setReadOnly(not editable)
                    elif hasattr(widget, 'setEnabled'):
                        widget.setEnabled(editable)
                else:
                    # Just using the table widget, set the cell text
                    cell=QtGui.QTableWidgetItem(str(value))
                    if editable:
                        cell.setFlags(Qt.ItemIsEditable | Qt.ItemIsEnabled)
                    else:
                        cell.setFlags(Qt.ItemIsEnabled)
                    twProperties.setItem(i, 2, cell)

                i+=1

    
    # callback indicating a cell in the table widget was changed
    def prop_cell_changed(self, twi):
        if twi.column() != 2: # the data column
            return

        row=twi.row()
        prop=str(self.window.twProperties.item(row, 1).text())
        self.property_changed(prop, str(twi.text()))

    # callback indicating one of the property widgets was modified
    def property_changed(self, prop, value=None):
        if not self.selected_shapebody[0]:
            return

        print('Trying to change %s to %s...' % (prop, value))
        shape, body=self.selected_shapebody
        for inst in (shape, body):
            if hasattr(inst, prop):
                try:
                    cur_value=getattr(inst, prop)
                    if isinstance(cur_value, b2Vec2):
                        m=re.search('\(?([\d\.]*)\s*,\s*([\d\.]*)\)?', value)
                        if m:
                            x, y=m.groups()
                            value=(float(x), float(y))
                except:
                    raise
                    pass

                try:
                    setattr(inst, prop, value)
                except:
                    print('Failed - %s' % sys.exc_info()[1])

    def ShowProperties(self, p):
        aabb = b2AABB(lowerBound=p-(0.001, 0.001), upperBound=p+(0.001, 0.001))

        # Query the world for overlapping shapes.
        query = fwQueryCallback(p)
        self.world.QueryAABB(query, aabb)
        
        if query.fixture:
            self.window.reset_properties_list()

            fixture=query.fixture
            body=fixture.body
            self._ShowProperties(body)

            shape=fixture.shape
            self._ShowProperties(shape)

            self.selected_shapebody=(shape, body)

    def Step(self, settings):
        super(Pyqt4Framework, self).Step(settings)

    def ConvertScreenToWorld(self, x, y):
        """
        PyQt4 gives us transformed positions, so no need to convert
        """
        return b2Vec2(x, y)
    
    DrawStringAt=lambda self, *args: self.renderer.DrawStringAt(*args)
    def Print(self, str, color=(229,153,153,255)):
        """
        Draw some text at the top status lines and advance to the next line.
        """
        self.DrawStringAt(5, self.textLine, str, color)
        self.textLine += self.renderer.font_spacing

    def Keyboard(self, key):
        """
        Callback indicating 'key' has been pressed down.
        The keys are mapped after pygame's style.

         from framework import Keys
         if key == Keys.K_z:
             ...
        """
        pass

    def KeyboardUp(self, key):
        """
        Callback indicating 'key' has been released.
        See Keyboard() for key information
        """
        pass

    def FixtureDestroyed(self, fixture):
        shape=fixture.shape
        if shape==self.selected_shapebody[0]:
            self.selected_shapebody=None, None
            self.window.reset_properties_list()
        if hash(shape) in self.renderer.item_cache:
            scene_items=self.renderer.item_cache[hash(shape)]
            for item in scene_items:
                self.window.scene.removeItem(item)
            del self.renderer.item_cache[hash(shape)]




import pygame
from pygame.locals import *
from .framework import *


try:
    from .pygame_gui import (fwGUI, gui)
    GUIEnabled = True
except:
    print('Unable to load PGU; menu disabled.')
    GUIEnabled = False

class PygameDraw(b2DrawExtended):
    """
    This debug draw class accepts callbacks from Box2D (which specifies what to draw)
    and handles all of the rendering.

    If you are writing your own game, you likely will not want to use debug drawing.
    Debug drawing, as its name implies, is for debugging.

    """

    surface = None
    axisScale = 10.0
    def __init__(self, **kwargs): 
        b2DrawExtended.__init__(self, **kwargs)
        self.flipX = False
        self.flipY = True
        self.convertVertices = True

    def StartDraw(self):
        self.zoom = 10
        self.center=(0,0)
        self.offset= (-900,550)
        self.screenSize=None

    def EndDraw(self): pass

    def DrawPoint(self, p, size, color):
        """
        Draw a single point at point p given a pixel size and color.
        """
        self.DrawCircle(p, size/self.zoom, color, drawwidth=0)
        
    def DrawAABB(self, aabb, color):
        """
        Draw a wireframe around the AABB with the given color.
        """
        points = [  (aabb.lowerBound.x, aabb.lowerBound.y ),
                    (aabb.upperBound.x, aabb.lowerBound.y ),
                    (aabb.upperBound.x, aabb.upperBound.y ),
                    (aabb.lowerBound.x, aabb.upperBound.y ) ]
        
        pygame.draw.aalines(self.surface, color, True, points)

    def DrawSegment(self, p1, p2, color):
        """
        Draw the line segment from p1-p2 with the specified color.
        """
        pygame.draw.aaline(self.surface, color.bytes, p1, p2)

    def DrawTransform(self, xf):
        """
        Draw the transform xf on the screen
        """
        p1 = xf.position
        p2 = self.to_screen(p1 + self.axisScale * xf.R.col1)
        p3 = self.to_screen(p1 + self.axisScale * xf.R.col2)
        p1 = self.to_screen(p1)

        pygame.draw.aaline(self.surface, (255,0,0), p1, p2)
        pygame.draw.aaline(self.surface, (0,255,0), p1, p3)

    def DrawCircle(self, center, radius, color, drawwidth=1):
        """
        Draw a wireframe circle given the center, radius, axis of orientation and color.
        """
        radius *= self.zoom
        if radius < 1: radius = 1
        else: radius = int(radius)

        pygame.draw.circle(self.surface, color.bytes, center, radius, drawwidth)

    def DrawSolidCircle(self, center, radius, axis, color):
        """
        Draw a solid circle given the center, radius, axis of orientation and color.
        """
        radius *= self.zoom
        if radius < 1: radius = 1
        else: radius = int(radius)

        pygame.draw.circle(self.surface, (color/2).bytes+[127], center, radius, 0)
        pygame.draw.circle(self.surface, color.bytes, center, radius, 1)
        pygame.draw.aaline(self.surface, (255,0,0), center, (center[0] - radius*axis[0], center[1] + radius*axis[1])) 

    def DrawPolygon(self, vertices, color):
        """
        Draw a wireframe polygon given the screen vertices with the specified color.
        """
        if not vertices:
            return

        if len(vertices) == 2:
            pygame.draw.aaline(self.surface, color.bytes, vertices[0], vertices)
        else:
            pygame.draw.polygon(self.surface, color.bytes, vertices, 1)
        
    def DrawSolidPolygon(self, vertices, color):
        """
        Draw a filled polygon given the screen vertices with the specified color.
        """
        if not vertices:
            return

        if len(vertices) == 2:
            pygame.draw.aaline(self.surface, color.bytes, vertices[0], vertices[1])
        else:
            pygame.draw.polygon(self.surface, (color/2).bytes+[127], vertices, 0)
            pygame.draw.polygon(self.surface, color.bytes, vertices, 1)

class PygameFramework(FrameworkBase):
    TEXTLINE_START=30
    def setup_keys(self):
        keys = [s for s in dir(pygame.locals) if s.startswith('K_')]
        for key in keys:
            value=getattr(pygame.locals, key)
            setattr(Keys, key, value)

    def __reset(self):
        # Screen/rendering-related
        self._viewZoom          = 10
        self._viewCenter        = 0
        self._viewOffset        = 0
        self.screenSize         = None
        self.rMouseDown         = False
        self.textLine           = 30
        self.font               = None
        self.fps                = 0


        # GUI-related (PGU)
        self.gui_app  =None
        self.gui_table=None
        self.setup_keys()
        
    def __init__(self):
        super(PygameFramework, self).__init__()

        if fwSettings.onlyInit: # testing mode doesn't initialize pygame
            return

        self.__reset()
        print('Initializing pygame framework...')
        pygame.init()
        caption= "Python Box2D Testbed - " + self.name
        pygame.display.set_caption(caption)
        self.screen = pygame.display.set_mode((1920,1080))
        self.screenSize = b2Vec2(*self.screen.get_size())
        #self.renderer = PygameDraw(surface=self.screen, test=self)
        self.renderer = PygameDraw(surface=self.screen)
        self.world.renderer=self.renderer
        
        try:
            self.font = pygame.font.Font(None, 24)
        except IOError:
            try:
                self.font = pygame.font.Font("freesansbold.ttf", 24)
            except IOError:
                print("Unable to load default font or 'freesansbold.ttf'")
                print("Disabling text drawing.")
                self.Print = lambda *args: 0
                self.DrawStringAt = lambda *args: 0

        # GUI Initialization
        if GUIEnabled:
            self.gui_app = gui.App()
            self.gui_table=fwGUI(self.settings)
            container = gui.Container(align=1,valign=-1)
            container.add(self.gui_table,0,0)
            self.gui_app.init(container)

        self.viewCenter = (0,0)
        self.groundbody = self.world.CreateBody()

    def setCenter(self, value):
        """
        Updates the view offset based on the center of the screen.
        
        Tells the debug draw to update its values also.
        """
        self._viewCenter = b2Vec2( *value )
        self._viewCenter *= self._viewZoom
        self._viewOffset = self._viewCenter - self.screenSize/2
    
    def setZoom(self, zoom):
        self._viewZoom = zoom

    viewZoom   = property(lambda self: self._viewZoom, setZoom,
                           doc='Zoom factor for the display')
    viewCenter = property(lambda self: self._viewCenter/self._viewZoom, setCenter, 
                           doc='Screen center in camera coordinates')
    viewOffset = property(lambda self: self._viewOffset,
                           doc='The offset of the top-left corner of the screen')

    def checkEvents(self):
        """
        Check for pygame events (mainly keyboard/mouse events).
        Passes the events onto the GUI also.
        """
        for event in pygame.event.get():
            if event.type == QUIT or (event.type == KEYDOWN and event.key == Keys.K_ESCAPE):
                return False
            elif event.type == KEYDOWN:
                self._Keyboard_Event(event.key, down=True)
            elif event.type == KEYUP:
                self._Keyboard_Event(event.key, down=False)
            elif event.type == MOUSEBUTTONDOWN:
                p = self.ConvertScreenToWorld(*event.pos)
                if event.button == 1: # left
                    mods = pygame.key.get_mods()
                    if mods & KMOD_LSHIFT:
                        self.ShiftMouseDown( p )
                    else:
                        self.MouseDown( p )
                elif event.button == 2: #middle
                    pass
                elif event.button == 3: #right
                    self.rMouseDown = True
                elif event.button == 4:
                    self.viewZoom *= 1.1
                elif event.button == 5:
                    self.viewZoom /= 1.1
            elif event.type == MOUSEBUTTONUP:
                p = self.ConvertScreenToWorld(*event.pos)
                if event.button == 3: #right
                    self.rMouseDown = False
                else:
                    self.MouseUp(p)
            elif event.type == MOUSEMOTION:
                p = self.ConvertScreenToWorld(*event.pos)

                self.MouseMove(p)

                if self.rMouseDown:
                    self.viewCenter -= (event.rel[0]/5.0, -event.rel[1]/5.0)

            if GUIEnabled:
                self.gui_app.event(event) #Pass the event to the GUI

        return True

    def run(self):
        """
        Main loop.

        Continues to run while checkEvents indicates the user has 
        requested to quit.

        Updates the screen and tells the GUI to paint itself.
        """

        # If any of the test constructors update the settings, reflect
        # those changes on the GUI before running
        if GUIEnabled:
            self.gui_table.updateGUI(self.settings)

        running = True
        clock = pygame.time.Clock()
        while running:
            running = self.checkEvents()
            self.screen.fill( (0,0,0) )
            # Check keys that should be checked every loop (not only on initial keydown)
            self.CheckKeys()

            # Run the simulation loop 
            self.SimulationLoop()

            if GUIEnabled and self.settings.drawMenu:
                self.gui_app.paint(self.screen)

            pygame.display.flip()
            clock.tick(self.settings.hz)
            self.fps = clock.get_fps()
    
        self.world.contactListener = None
        self.world.destructionListener=None
        self.world.renderer=None

    def _Keyboard_Event(self, key, down=True):
        """
        Internal keyboard event, don't override this.

        Checks for the initial keydown of the basic testbed keys. Passes the unused
        ones onto the test via the Keyboard() function.
        """
        if down:
            if key==Keys.K_z:       # Zoom in
                self.viewZoom = min(1.1 * self.viewZoom, 50.0)
            elif key==Keys.K_x:     # Zoom out
                self.viewZoom = max(0.9 * self.viewZoom, 0.02)
            elif key==Keys.K_SPACE: # Launch a bomb
                self.LaunchRandomBomb()
            elif key==Keys.K_F1:    # Toggle drawing the menu
                self.settings.drawMenu = not self.settings.drawMenu
            elif key==Keys.K_F2:    # Do a single step
                self.settings.singleStep=True
                if GUIEnabled:
                    self.gui_table.updateGUI(self.settings)
            else:              # Inform the test of the key press
                self.Keyboard(key)
        else:
            self.KeyboardUp(key)

    def CheckKeys(self):
        """
        Check the keys that are evaluated on every main loop iteration.
        I.e., they aren't just evaluated when first pressed down
        """

        pygame.event.pump()
        self.keys = keys = pygame.key.get_pressed()
        if keys[Keys.K_LEFT]:
            self.viewCenter -= (0, 0)
            self.viewCenter -= (0.5, 0)
        elif keys[Keys.K_RIGHT]:
            self.viewCenter += (0, 0)
            self.viewCenter += (0.5, 0)

        if keys[Keys.K_UP]:
            self.viewCenter += (0, 0)
            self.viewCenter += (0, 0.5)
        elif keys[Keys.K_DOWN]:
            self.viewCenter -= (0, 0)
            self.viewCenter -= (0, 0.5)

        if keys[Keys.K_HOME]:
            self.viewZoom = 1.0
            self.viewCenter = (0.0, 20.0)

   
    def Step(self, settings):
        if GUIEnabled:
            # Update the settings based on the GUI
            self.gui_table.updateSettings(self.settings)

        super(PygameFramework, self).Step(settings)

        if GUIEnabled:
            # In case during the step the settings changed, update the GUI reflecting
            # those settings.
            self.gui_table.updateGUI(self.settings)

    def ConvertScreenToWorld(self, x, y):
        return b2Vec2((x + self.viewOffset.x) / self.viewZoom,
                           ((self.screenSize.y - y + self.viewOffset.y) / self.viewZoom))

    def DrawStringAt(self, x, y, str, color=(229,153,153,255)):
        """
        Draw some text, str, at screen coordinates (x, y).
        """
        self.screen.blit(self.font.render(str, True, color), (x,y))

    def Print(self, str, color=(229,153,153,255)):
        """
        Draw some text at the top status lines
        and advance to the next line.
        """
        self.screen.blit(self.font.render(str, True, color), (15,self.textLine-10))
        self.textLine += 15

    def Keyboard(self, key):
        """
        Callback indicating 'key' has been pressed down.
        The keys are mapped after pygame's style.

         from framework import Keys
         if key == Keys.K_z:
             ...
        """
        pass

    def KeyboardUp(self, key):
        """
        Callback indicating 'key' has been released.
        See Keyboard() for key information
        """
        pass


import os, getopt  # For system operations and argument parsing
import re
from .wsWordObj import *
from .NGrams import wsNGrams
import sys
sys.path.append('./framework/')
sys.path.append('./postprocessing/frames/')
from .framework.framework import Framework
from .framework.framework import *
from pygame.locals import * # Some parts of the PyGame framework to import locally




class WordSwarm(Framework):
    # Settings
    name = ""
    saveFrames = True
    csvName = 'matrix.csv'
    saveFolder = './postprocessing/frames/'

    frameN = 0;
    shapes = [];
    fixtures = [];
    bodies = [];
    joints = [];
    wordObjs = []
    sun = []
    wordHue = (-1, -1)
    topN = sys.maxsize

    # Size of words (m)
    maxSize = 6
    minSize = 0.1

    # Sun strength
    frequency = 0.1
    damping = 2

    def __init__(self, argv):

        # Initialize pygame/pybox2d framework/wcd ..orld
        super(WordSwarm, self).__init__()
        self.world.gravity = (0, 0);
        startDateStr = None
        endDateStr = None

        # Parse arguments
        try:
            opts, args = getopt.getopt(argv,
                                       "hst:i:d:m:c:b:e:n:", ["ifile="])
        except getopt.GetoptError:
            print
            'Invalid argument'
            print(argv)
            print('try:')
            print
            'WordSwarm.py -n <topN] -s -t <title> -i <inputfile> -d <outputFolder> -m <maxWordHeight> -c <HexHue1_HexHue2> -b <startDate YYYYMMDD> -e <endDate YYYYMMDD>'
            sys.exit(2)

        for opt, arg in opts:
            if opt == '-h':
                print
                'WordSwarm.py -n <topN] -s -t <title> -i <inputfile> -d <outputFolder> -m <maxWordHeight> -c <HexHue1_HexHue2> -b <startDate YYYYMMDD> -e <endDate YYYYMMDD>'
                sys.exit()
            elif opt in ("-i", "--ifile"):

                print('Reading csv: %s' % arg)
                self.csvName = arg

            elif opt in ("-d"):
                self.saveFrames = True
                self.saveFolder = arg
                print('Saving frames to %s' % self.saveFolder)

            elif opt in ("-s"):
                self.saveFrames = True
                print('Saving frames to %s' % self.saveFolder)

            elif opt in ("-t"):

                print('WordSwarm title: %s' % arg)
                self.name = arg

            elif opt in ("-m"):
                self.maxSize = int(arg)
                print('Max word height set to: %d' % self.maxSize)

            elif opt in ("-c"):
                self.wordHue = (int(arg[0:2], 16), int(arg[3:5], 16))
                print('Words will have hues %d or %d' % (self.wordHue[0], self.wordHue[1]))

            elif opt in ("-b"):
                startDateStr = arg
                print('Starting animation at %s' % startDateStr)

            elif opt in ("-e"):
                endDateStr = arg
                print('Ending animation on %s' % endDateStr)

            elif opt in ("-n"):
                self.topN = int(arg)
                print('Displaying only the first %d results in csv' % self.topN)

        # Create output directory if required
        if self.saveFrames:
            if not os.path.exists(self.saveFolder):
                os.makedirs(self.saveFolder)
            else:
                self.purge(self.saveFolder, '.*')

        # Create ngrams database
        self.nGrams = wsNGrams(self.csvName, startDateStr, endDateStr, self.topN)

        box_half_size = (0.5, 0.5)

        # Place word objects
        self.shapes = [None] * self.nGrams.nWords
        self.fixtures = [None] * self.nGrams.nWords
        self.bodies = [None] * self.nGrams.nWords
        self.joints = [None] * self.nGrams.nWords
        self.wordObjs = [None] * self.nGrams.nWords;
        for word_k in range(0, self.nGrams.nWords):

            # The centre of the universe
            self.sun.append(self.world.CreateStaticBody(
                position=b2Vec2(random.uniform(
                    -45, 65), 0)));

            # Create word object
            if self.nGrams.areColors == 'hue':
                self.wordObjs[word_k] = wsWordObj(
                    self.nGrams.words[word_k], [self.nGrams.colors[word_k]])
            elif self.nGrams.areColors == 'rgb':
                self.wordObjs[word_k] = wsWordObj(
                    self.nGrams.words[word_k], self.nGrams.colors[word_k])
            else:
                self.wordObjs[word_k] = wsWordObj(
                    self.nGrams.words[word_k], self.wordHue)

            # Create body for word
            self.bodies[word_k] = self.world.CreateDynamicBody(
                position=(random.uniform(-60, 70), random.uniform(-40, 40)))

            # Add fixture to body
            self.fixtures[word_k] = self.bodies[word_k].CreateFixture(
                b2FixtureDef(shape=b2PolygonShape(box=(
                    box_half_size[0] / self.wordObjs[word_k].paddedAR, box_half_size[0]))))

            # Link word object to sun
            self.joints[word_k] = self.world.CreateJoint(
                b2DistanceJointDef(
                    frequencyHz=self.frequency,
                    dampingRatio=self.damping,
                    bodyA=self.sun[word_k],
                    bodyB=self.bodies[word_k],
                    localAnchorA=(0, 0),
                    localAnchorB=(0, 0),
                    length=0.5))

    # Removes files from a directory matching a pattern
    def purge(self, dir, pattern):
        for f in os.listdir(dir):
            if re.search(pattern, f):
                os.remove(os.path.join(dir, f))

    # Converts a size in (m) to a size in (px)
    def convertWorld2Screen(self, size_m):

        # Generate faux coordinates in (m)
        coord_m1 = size_m
        coord_m2 = list(size_m)
        coord_m2[0] = -coord_m2[0]
        coord_m2[1] = -coord_m2[1]

        # Convert faux coordinates to (px)
        coord_px1 = self.renderer.to_screen(coord_m1)
        coord_px2 = self.renderer.to_screen(coord_m2)

        # Calculate size from new coordinates
        size_p = [abs(coord_px1[0] - coord_px2[0]),
                  abs(coord_px1[1] - coord_px2[1])]

        return size_p

    # Draw the date progress bar
    # @TODO Scale text and line weight with screen size
    def Draw_Date(self, date_k):
        months = ['', 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

        color = (255, 255, 255)
        dateTxt = freetype.Font(None)
        dateTxt.size = 24
        dateTxt.fgcolor = color

        top = (int(self.screen.get_height() * 0.075), int(self.screen.get_height() * 0.125))
        #top = (250, 180)
        #bot = (250, 910)
        bot = (int(self.screen.get_height() * 0.075), int(self.screen.get_height() * (1 - 0.125)))

        self.screen.blit(dateTxt.render(self.name)[0],
                         (self.screen.get_height() * 0.03, self.screen.get_height() * 0.03))

        pygame.draw.line(self.screen, color, top, bot, 4)
        pygame.draw.line(self.screen, color, (top[0] - 4, top[1]), (top[0] + 4, top[1]), 4)
        pygame.draw.line(self.screen, color, (bot[0] - 4, bot[1]), (bot[0] + 4, bot[1]), 4)

        self.screen.blit(dateTxt.render('%s %04d' % (
            months[self.nGrams.dates[0].month],
            self.nGrams.dates[0].year))[0],
                         (self.screen.get_height() * 0.03, top[1] - 40))

        self.screen.blit(dateTxt.render('%s %04d' % (
            months[self.nGrams.dates[-1].month],
            self.nGrams.dates[-1].year))[0],
                         (self.screen.get_height() * 0.03, bot[1] + 20))

        progress = ((self.nGrams.dates[date_k] - self.nGrams.dates[0]).total_seconds() /
                    (self.nGrams.dates[-1] - self.nGrams.dates[0]).total_seconds())

        pos = (top[0] + 1, int((bot[1] - top[1]) * progress + top[1]))
        pygame.draw.circle(self.screen, color, pos, 8)

        self.screen.blit(dateTxt.render('%s %04d' % (
            months[self.nGrams.dates[date_k].month],
            self.nGrams.dates[date_k].year))[0],
                         (pos[0] + 20, pos[1] - 10))


    def Step(self, settings):
        self.screen.fill((0, 0, 0))
        self.frameN = self.frameN + 1;

        # Update ngram date every n-frames
        nFrames = 90
        date_k = int(self.frameN / nFrames) + 1  # Date moving towards
        phase = (self.frameN % nFrames) / float(nFrames)  # (0-1) way to new date

        # Stop if at the end of the dataset
        if date_k == self.nGrams.nDates:
            print('WordSwarm animation complete')
            print('Hope you enjoyed the show!')
            exit()

        dateTxt = freetype.Font(None)
        dateTxt.size = 10
        dateTxt.fgcolor = (255, 255, 255)

        self.Draw_Date(date_k)  # Draw the date

        # Update size of each word
        # @TODO There is a memory leak in creating the new bodies (it doesn't delete the old ones)
        for word_k in range(0, self.nGrams.nWords):

            # Calculate word sizes
            if phase == 1:
                newSize = (0, 0)

                newSize[1] = (self.maxSize - self.minSize) * (
                            self.nGrams.counts[word_k][date_k - 1] / float(self.nGrams.maxCount)) + self.minSize

                newSize[0] = newSize[1] / self.wordObjs[word_k].paddedAR

                self.wordObjs[word_k].boxSize = self.convertWorld2Screen(newSize)
            else:
                newSize = list((0, 0))
                newSize[1] = (self.maxSize - self.minSize) * (
                            self.nGrams.counts[word_k][date_k] / float(self.nGrams.maxCount)) + self.minSize
                newSize[0] = newSize[1] / self.wordObjs[word_k].paddedAR

                oldSize = list((0, 0))
                oldSize[1] = (self.maxSize - self.minSize) * (
                            self.nGrams.counts[word_k][date_k - wsNGrams.nDates] / float(self.nGrams.maxCount)) + self.minSize
                oldSize[0] = oldSize[1] / self.wordObjs[word_k].paddedAR

                size = list((0, 0))
                size[0] = (newSize[0] - oldSize[0]) * phase + oldSize[0]
                size[1] = (newSize[1] - oldSize[1]) * phase + oldSize[1]

                self.wordObjs[word_k].boxSize = self.convertWorld2Screen(size)

            # Re-create fixture
            self.bodies[word_k].DestroyFixture(self.fixtures[word_k])
            self.fixtures[word_k] = None
            self.fixtures[word_k] = self.bodies[word_k].CreateFixture(
                b2FixtureDef(shape=b2PolygonShape(box=newSize)))

            # Redraw word in new shape
            pos = self.renderer.to_screen(self.bodies[word_k].position);
            self.wordObjs[word_k].Draw(self.screen, pos)

        # Save frames to create film
        if self.saveFrames:
            pygame.image.save(self.screen,
                              self.saveFolder + int(self.frameN).__format__('')
                              + '.png')

        Framework.Step(self, settings);




class fwSettings(object):
    backend='pygame'        # The default backend to use in (can be: pyglet, pygame, etc.)

    # Physics options
    hz=60.0
    velocityIterations=8
    positionIterations=3
    enableWarmStarting=True   # Makes physics results more accurate (see Box2D wiki)
    enableContinuous=True     # Calculate time of impact
    enableSubStepping=False
    
    # Drawing
    drawStats=False
    drawShapes=False
    drawJoints=False
    drawCoreShapes=False
    drawAABBs=False
    drawOBBs=False
    drawPairs=False
    drawContactPoints=False
    maxContactPoints=100
    drawContactNormals=False
    drawFPS=False
    drawMenu=False             # toggle by pressing F1
    drawCOMs=False            # Centers of mass
    pointSize=2.5             # pixel radius for drawing points

    # Miscellaneous testbed options
    pause=False
    singleStep=False
    onlyInit=False            # run the test's initialization without graphics, and then quit (for testing)

#             text                  variable
checkboxes =( ("Warm Starting"   , "enableWarmStarting"), 
              ("Time of Impact"  , "enableContinuous"), 
              ("Sub-Stepping"    , "enableSubStepping"),
              ("Draw"            , None),
              ("Shapes"          , "drawShapes"), 
              ("Joints"          , "drawJoints"), 
              ("AABBs"           , "drawAABBs"), 
              ("Pairs"           , "drawPairs"), 
              ("Contact Points"  , "drawContactPoints"), 
              ("Contact Normals" , "drawContactNormals"), 
              ("Center of Masses", "drawCOMs"), 
              ("Statistics"      , "drawStats"),
              ("FPS"             , "drawFPS"),
              ("Control"         , None),
              ("Pause"           , "pause"),
              ("Single Step"     , "singleStep") )

sliders = [
    { 'name' : 'hz'                , 'text' : 'Hertz'    , 'min' : 5, 'max' : 200 },
    { 'name' : 'positionIterations', 'text' : 'Pos Iters', 'min' : 0, 'max' : 100 },
    { 'name' : 'velocityIterations', 'text' : 'Vel Iters', 'min' : 1, 'max' : 500 },
]






import setuptools

with open("README.md", "r") as fh:
    description = fh.read()

    setuptools.setup(
        name="AnimatedWordCloud",
        version="1.0.4",
        author="Petr Koráb",
        author_email="xpetrkorab@gmail.com",
        packages=["AnimatedWordCloud"],
        description="Animated version of classic word cloud for time-series text data",
        long_description=description,
        long_description_content_type="text/markdown",
        url="https://github.com/PetrKorab/AnimatedWordCloud",
        include_package_data = True,
        python_requires='>=3.8, <3.9',
        install_requires = ['arabica == 1.7.6',
                            'Box2D == 2.3.10',
                            'chardet == 3.0.4',
                            'beautifulsoup4 == 4.12.2',
                            'openpyxl == 3.1.2',
                            'pygame == 2.5.0',
                            'PyQt6 == 6.5.2',
                            'PyQt6-Qt6 == 6.5.2',
                            'PyQt6-sip == 13.5.2',
                            'SQLAlchemy == 1.3.2',
                            'ftfy == 6.1.1',
                            'pygame-pgu == 0.21'],
        license='OSI Approved :: Apache Software License'
    )


from .preprocessing import prep
from .wsWordObj_monthly import *
from .wsWordObj import *
sys.path.append('./framework/')
from .framework.framework import main


class animated_word_cloud():
    def __init__(self, text, time, date_format, freq, ngram, stopwords):
        self.text = text
        self.time = time
        self.date_format = date_format
        self.freq = freq
        self.ngram = ngram
        self.stopwords = stopwords
  
        output = prep(text_prep=text, time=time, date_format=date_format, ngram=ngram, freq=freq, stopwords=stopwords)
        output.to_csv("matrix.csv", index=False, encoding="utf8")
        print("Data matrix created ...")

        if freq == "Y":
            from .WordSwarm import WordSwarm
            main(WordSwarm, sys.argv[1:])
                
        elif freq == "M":
            from .WordSwarm_monthly import WordSwarm
            main(WordSwarm, sys.argv[1:])

        else:
            print(""""Incorrect frequency specification Use "Y" or "M" """)



from ftfy import fix_encoding


def prep(text_prep: str,
        time: str,
        date_format: str,
        ngram: str,
        freq:str,
        stopwords: []):

        import pandas as pd
        from arabica import arabica_freq

        global matrix
        if ngram == 1:
            cool=pd.DataFrame()
            cool = cool.dropna()
            cool['text']=text_prep
            text = text_prep
            cool['time']=time
            cool['time']=cool['time'].astype(str)
            cool['time'] = pd.to_datetime(cool['time'], errors= 'coerce')
            if freq == "Y":
                df = arabica_freq(text = text,
                                  time = time,
                                  date_format = date_format,
                                  time_freq = 'Y',
                                  max_words = 50,
                                  stopwords = stopwords,
                                  skip = None,
                                  numbers = True,
                                  lower_case = True)

                unigram = df[["period", "unigram"]]
                values = unigram["unigram"]
                period = unigram["period"]
                values = values.str.split(pat=',',expand=True)
                period = period.astype(str)
                test = pd.concat([period, values], axis=1)

                colnames = []

                for name in test.columns:
                    name = str(name)
                    name = "word" + name
                    colnames.append(name)

                test.columns = colnames
                test["period"] = test["wordperiod"].rename("period", inplace = True)
                test=test.iloc[:,1:]
                l = test.melt(id_vars = "period")
                l = l[["period","value"]]
                l = test.melt(id_vars = "period")
                l = l[["period","value"]]
                l.columns = ["period", "word"]
                freq = l["word"].str.split(pat = ":", expand = True)
                freq.columns = ["word", "freq"]
                period = l["period"]
                freq.columns = ["word", "freq"]
                pokus = pd.concat([period, freq],axis=1)
                df = pokus.reset_index().pivot_table(values="freq", index="word", columns="period", aggfunc='mean')
                df.reset_index(inplace=True)
                df = df[df['word'] != 'NaT']
                df = df.fillna(0)
                df.rename(columns={df.columns[0]: ' '}, inplace=True)
                df.iloc[:,0] = df.iloc[:,0].apply(fix_encoding)
                total_sum = df.iloc[1:, 1:].sum().sum()
                if total_sum > 15000:
                    # Calculate the scaling factor
                    scaling_factor = 15000 / total_sum
                    # Divide every element in the DataFrame by the scaling factor
                    df.iloc[1:, 1:] = df.iloc[1:, 1:] * scaling_factor


            elif freq == "M":
                df = arabica_freq(text=text,
                                  time=time,
                                  date_format=date_format,
                                  time_freq='M',
                                  max_words=70,
                                  stopwords=stopwords,
                                  skip=None,
                                  numbers=True,
                                  lower_case=True)

                unigram = df[["period", "unigram"]]
                values = unigram["unigram"]
                period = unigram["period"]
                values = values.str.split(pat=',', expand=True)
                period = period.astype(str)
                test = pd.concat([period, values], axis=1)

                colnames = []

                for name in test.columns:
                    name = str(name)
                    name = "word" + name
                    colnames.append(name)

                test.columns = colnames
                test["period"] = test["wordperiod"].rename("period", inplace=True)
                test = test.dropna()
                test = test.iloc[:, 1:]
                l = test.melt(id_vars="period")
                l = l[["period", "value"]]
                l = test.melt(id_vars="period")
                l = l[["period", "value"]]
                l.columns = ["period", "word"]
                freq = l["word"].str.split(pat=":", expand=True)
                freq.columns = ["word", "freq"]
                print("freq")
                period = l["period"]
                freq.columns = ["word", "freq"]
                pokus = pd.concat([period, freq], axis=1)
                df = pokus.reset_index().pivot_table(values="freq", index="word", columns="period", aggfunc='mean')
                df.reset_index(inplace=True)
                df = df[df['word'] != 'NaT']
                df = df.fillna(0)
                df.rename(columns={df.columns[0]: ' '}, inplace=True)
                df.iloc[:,0] = df.iloc[:,0].apply(fix_encoding)
                total_sum = df.iloc[1:, 1:].sum().sum()
                if total_sum > 15000:
                    # Calculate the scaling factor
                    scaling_factor = 15000 / total_sum
                    # Divide every element in the DataFrame by the scaling factor
                    df.iloc[1:, 1:] = df.iloc[1:, 1:] * scaling_factor

        return df